import inspect
from enum import Enum
from functools import reduce
from itertools import chain
from typing import Any, Dict, List, Optional, no_type_check, Union
from typing_extensions import get_origin, get_args

from django.core.serializers.json import DjangoJSONEncoder
from django.db.models import Manager, Model
from django.db.models.fields.files import ImageFieldFile
from django.db.models.fields.reverse_related import ForeignObjectRel, OneToOneRel
from django.utils.encoding import force_str
from django.utils.functional import Promise
from pydantic import BaseModel, create_model
from pydantic.errors import PydanticUserError
from pydantic._internal._model_construction import ModelMetaclass


from .fields import ModelSchemaField

_is_base_model_class_defined = False


class ModelSchemaJSONEncoder(DjangoJSONEncoder):
    @no_type_check
    def default(self, obj):  # pragma: nocover
        if isinstance(obj, Promise):
            return force_str(obj)

        return super().default(obj)


def get_field_name(field) -> str:
    if issubclass(field.__class__, ForeignObjectRel) and not issubclass(
        field.__class__, OneToOneRel
    ):
        return getattr(field, "related_name", None) or f"{field.name}_set"
    else:
        return getattr(field, "name", field)


class ModelSchemaMetaclass(ModelMetaclass):
    @no_type_check
    def __new__(mcs, name: str, bases: tuple, namespace: dict, **kwargs):
        cls = super().__new__(mcs, name, bases, namespace, **kwargs)
        for base in reversed(bases):
            if (
                _is_base_model_class_defined
                and issubclass(base, ModelSchema)
                and base == ModelSchema
            ):

                config = namespace["model_config"]
                include = config.get("include", None)
                exclude = config.get("exclude", None)

                if include and exclude:
                    raise PydanticUserError(
                        "Only one of 'include' or 'exclude' should be set in "
                        "configuration.",
                        code="include-exclude-mutually-exclusive",
                    )

                annotations = namespace.get("__annotations__", {})

                try:
                    fields = config["model"]._meta.get_fields()
                except (AttributeError, KeyError) as exc:
                    raise PydanticUserError(
                        f'{exc} (Is `model_config["model"]` a valid Django model class?)',
                        code="class-not-valid",
                    )

                if include == "__annotations__":
                    include = list(annotations.keys())
                    cls.model_config["include"] = include
                elif include is None and exclude is None:
                    include = list(annotations.keys()) + [
                        get_field_name(f) for f in fields
                    ]
                    cls.model_config["include"] = include

                field_values = {}
                _seen = set()

                for field in chain(fields, annotations.copy()):
                    field_name = get_field_name(field)

                    if (
                        field_name in _seen
                        or (include and field_name not in include)
                        or (exclude and field_name in exclude)
                    ):
                        continue

                    _seen.add(field_name)

                    python_type = None
                    pydantic_field = None
                    if field_name in annotations and field_name in namespace:

                        python_type = annotations.pop(field_name)
                        pydantic_field = namespace[field_name]
                        if (
                            hasattr(pydantic_field, "default_factory")
                            and pydantic_field.default_factory
                        ):
                            pydantic_field = pydantic_field.default_factory()

                    elif field_name in annotations:
                        python_type = annotations.pop(field_name)
                        pydantic_field = (
                            None if Optional[python_type] == python_type else Ellipsis
                        )

                    else:
                        python_type, pydantic_field = ModelSchemaField(field, name)

                    field_values[field_name] = (python_type, pydantic_field)

                cls.__doc__ = namespace.get("__doc__", config["model"].__doc__)
                cls.model_fields = {}
                cls.__alias_map__ = {
                    getattr(model_field[1], "alias", None) or field_name: field_name
                    for field_name, model_field in field_values.items()
                }
                model_schema = create_model(
                    name,
                    __base__=cls,
                    __module__=cls.__module__,
                    __doc__=cls.__doc__,
                    **field_values,
                )
                return model_schema

        return cls


class ProxyGetterNestedObj:
    def __init__(self, obj: Any, schema_class):
        self._obj = obj
        self.schema_class = schema_class

    def get(self, key: Any, default: Any = None) -> Any:
        if "__" in key:
            # Allow double underscores aliases: `first_name: str = Field(alias="user__first_name")`
            keys_map = key.split("__")
            attr = reduce(lambda a, b: getattr(a, b, default), keys_map, self._obj)
        else:
            attr = getattr(self._obj, key, None)
        is_manager = issubclass(attr.__class__, Manager)

        alias = self.schema_class.__alias_map__[key]
        field = self.schema_class.model_fields[alias]

        typing_args = get_args(field.annotation)
        typing_origin = get_origin(field.annotation)
        if typing_origin == Union:
            outer_type_ = typing_args[0]
        elif typing_origin is None:
            outer_type_ = field.annotation
        # TODO do we have test for a list thingie?
        elif typing_args[0] == List:
            outer_type_ = typing_origin
        elif is_manager and field.annotation == List[Dict[str, int]]:
            outer_type_ = field.annotation
        else:
            outer_type_ = typing_args[0]

        if is_manager and outer_type_ == List[Dict[str, int]]:
            attr = list(attr.all().values("id"))
        elif is_manager:
            attr = list(attr.all())
        elif outer_type_ == int and issubclass(type(attr), Model):
            attr = attr.id
        elif inspect.isclass(type(attr)) and issubclass(type(attr), Enum):
            attr = attr.value
        elif issubclass(attr.__class__, ImageFieldFile) and issubclass(
            outer_type_, str
        ):
            attr = attr.name
        return attr

    def _get_annotation_objects(self, value, annotation):
        """Value is an object, and the value needs to resolve into a dict"""
        if isinstance(value, list):
            return [ProxyGetterNestedObj(o, annotation).dict() for o in value]
        return ProxyGetterNestedObj(value, annotation).dict()

    def dict(self) -> dict:
        """
        Might not be needed with "from_attributes=True" in model_config
        """
        fields = self.schema_class.model_fields
        data = {}
        for key, fieldinfo in fields.items():
            annotation = fieldinfo.annotation
            if get_origin(annotation) == list:
                # Pick the underlying annotation
                annotation = get_args(annotation)[0]

            if inspect.isclass(annotation) and issubclass(annotation, ModelSchema):
                data[key] = self._get_annotation_objects(self.get(key), annotation)
            else:
                key = fieldinfo.alias if fieldinfo.alias else key
                data[key] = self.get(key)
        return data


class ModelSchema(BaseModel, metaclass=ModelSchemaMetaclass):

    def __eq__(self, other: Any) -> bool:
        result = super().__eq__(other)
        if isinstance(result, bool):
            return result

        if result is NotImplemented and isinstance(other, dict):
            return self.model_dump() == other
        return result

    @classmethod
    def model_json_schema(cls, *args, **kwargs):
        result = super().model_json_schema(*args, **kwargs)
        if cls.model_config.get("include"):
            include = cls.model_config.get("include", [])
            for key in list(result["properties"].keys()):
                if key not in include:
                    del result["properties"][key]

        if cls.model_config.get("exclude"):
            exclude = cls.model_config.get("exclude")
            for key in list(result["properties"].keys()):
                if key in exclude:
                    del result["properties"][key]

        return result

    @classmethod
    @no_type_check
    def get_field_names(cls) -> List[str]:
        if cls.model_config.get("exclude"):
            django_model_fields = cls.model_config["model"]._meta.get_fields()
            all_fields = [f.name for f in django_model_fields]
            return [
                name for name in all_fields if name not in cls.model_config["exclude"]
            ]
        return cls.model_config.get("include", [])

    @classmethod
    def from_orm(cls, *args, **kwargs):
        return cls.from_django(*args, **kwargs)

    @classmethod
    def from_django(cls, objs, many=False, context={}):
        if many:
            result_objs = []
            for obj in objs:
                obj = ProxyGetterNestedObj(obj, cls)
                instance = cls(**obj.dict())
                result_objs.append(cls.model_validate(instance, context=context))
            return result_objs

        obj = ProxyGetterNestedObj(objs, cls)
        instance = cls(**obj.dict())
        return cls.model_validate(instance, context=context)


_is_base_model_class_defined = True



import logging
from datetime import date, datetime, time, timedelta
from decimal import Decimal
from enum import Enum
from typing import Any, Dict, List, Union, Optional
import typing
from uuid import UUID

from django.utils.functional import Promise
from pydantic import IPvAnyAddress, Json
from pydantic.fields import FieldInfo
from pydantic.v1.fields import Required
from pydantic_core import PydanticUndefined

try:
    from django.db.models.fields import NOT_PROVIDED
except ImportError:

    class NOT_PROVIDED:
        pass


logger = logging.getLogger("djantic")


INT_TYPES = [
    "AutoField",
    "BigAutoField",
    "IntegerField",
    "SmallIntegerField",
    "BigIntegerField",
    "PositiveIntegerField",
    "PositiveSmallIntegerField",
]

STR_TYPES = [
    "CharField",
    "EmailField",
    "URLField",
    "SlugField",
    "TextField",
    "FilePathField",
    "FileField",
]


FIELD_TYPES = {
    "GenericIPAddressField": IPvAnyAddress,
    "BooleanField": bool,
    "BinaryField": bytes,
    "DateField": date,
    "DateTimeField": datetime,
    "DurationField": timedelta,
    "TimeField": time,
    "DecimalField": Decimal,
    "FloatField": float,
    "UUIDField": UUID,
    "JSONField": Union[Json, dict, list],  # TODO: Configure this using default
    "ArrayField": List,
    # "IntegerRangeField",
    # "BigIntegerRangeField",
    # "CICharField",
    # "CIEmailField",
    # "CIText",
    # "CITextField",
    # "DateRangeField",
    # "DateTimeRangeField",
    # "DecimalRangeField",
    # "FloatRangeField",
    # "HStoreField",
    # "RangeBoundary",
    # "RangeField",
    # "RangeOperators",
}


def get_internal_type(field):
    python_type = None
    internal_type = field.get_internal_type()
    if internal_type in STR_TYPES:
        python_type = str

    elif internal_type in INT_TYPES:
        python_type = int

    elif internal_type in FIELD_TYPES:
        python_type = FIELD_TYPES[internal_type]

    else:  # pragma: nocover
        for field_class in type(field).__mro__:
            get_internal_type = getattr(field_class, "get_internal_type", None)
            if get_internal_type:
                _internal_type = get_internal_type(field_class())
                if _internal_type in FIELD_TYPES:
                    python_type = FIELD_TYPES[_internal_type]
                    break
    return python_type


def ModelSchemaField(field: Any, schema_name: str) -> tuple:
    default = Required
    default_factory = None
    description = None
    title = None
    max_length = None

    try:
        deconstructed = field.deconstruct()
        field_options = deconstructed[3] or {}
        blank = field_options.pop("blank", False)
        null = field_options.pop("null", False)
    except AttributeError:
        # Field does not have `deconstruct` method, just default to false
        blank = False
        null = False

    if field.is_relation:
        if not field.related_model:
            internal_type = field.model._meta.pk.get_internal_type()
        else:
            internal_type = field.related_model._meta.pk.get_internal_type()
            if not field.concrete and field.auto_created or field.null:
                default = None

        pk_type = FIELD_TYPES.get(internal_type, int)
        if field.one_to_many or field.many_to_many:
            python_type = List[Dict[str, pk_type]]
        else:
            python_type = pk_type

        if blank or null:
            python_type = Union[python_type, None]

    else:
        if field.choices:
            enum_choices = {}
            for k, v in field.choices:
                if Promise in type(v).__mro__:
                    v = str(v)
                enum_choices[v] = k
            if field.blank:
                enum_choices["_blank"] = ""

            enum_prefix = (
                f"{schema_name.replace('_', '')}{field.name.title().replace('_', '')}"
            )
            python_type = Enum(  # type: ignore
                f"{enum_prefix}Enum",
                enum_choices,
                module=__name__,
            )

            if field.has_default() and isinstance(field.default, Enum):
                default = field.default.value
        else:
            python_type = get_internal_type(field)
            if field.get_internal_type() == "ArrayField":
                python_type = List[get_internal_type(field.base_field)]

        if python_type is None:
            logger.warning(
                "%s is currently unhandled, defaulting to str.", field.__class__
            )
            python_type = str

        if default is Required and field.has_default():
            if callable(field.default):
                default_factory = field.default
                default = PydanticUndefined
            else:
                default = field.default
        elif field.primary_key or blank or null:
            default = None
            python_type = Union[python_type, None]

        if (default is not None or NOT_PROVIDED) and field.null:
            python_type = Union[python_type, None]

        description = field.help_text
        title = field.verbose_name.title()

    if not description:
        related_field_name = None
        if field.related_model:
            related_field_name = field.target_field.name

        description = related_field_name or field.name

    if (
        getattr(field, "get_internal_type", lambda: None)() in STR_TYPES
        and not field.choices
    ):
        max_length = field.max_length

    field_info = FieldInfo(
        default=default,
        default_factory=default_factory,
        title=title,
        description=str(description),
        max_length=max_length,
    )

    field_is_optional = all([
        getattr(field, "null", None),
        field.is_relation,
        # A list that is null, is the empty list. So there is no need
        # to make it nullable.
        typing.get_origin(python_type) is not list
    ])
    if field_is_optional:
        python_type = Optional[python_type]

    return (
        python_type,
        field_info
    )
